;;;; fp-utils.scm

(module fp-utils

(;export
  fpzero?
  fppositive?
  fpnegative?
  fpeven?
  fpodd?
  ;
  fpmodulo
  fpquotient
  fpremainder
  ;
  fpfraction
  ;
  fptruncate/precision
  fpround/precision
  fpceiling/precision
  fpfloor/precision
  ;
  fp~=
  ;
  fpsqr fpcub
  ;
  fpprecision-factor
  ;
  fpdegree->radian fpradian->degree
  ;
  fpdistance fpdistance*)

(import scheme chicken foreign)

(declare
  (bound-to-procedure
    ##sys#flonum-fraction
    ##sys#check-inexact) )

;;;

;;

(define C_fmod
  (foreign-lambda double "fmod" double double))

(define C_remainder
  (foreign-lambda double "remainder" double double))

(define (*fpeven? n)
  (fp= 0.0 (##sys#flonum-fraction (fp/ n 2.0))) )

;;

(define (fpzero? n)
  (fp= 0.0 n) )

(define (fppositive? n)
  (fp< 0.0 n) )

(define (fpnegative? n)
  (fp> 0.0 n) )

(define (fpeven? n)
  (and
    (fpinteger? n)
    (*fpeven? n)) )

(define (fpodd? n)
  (not (fpeven? n)) )

;;

(define (fpmodulo x y)
  (##sys#check-inexact x 'fpmodulo)
  (##sys#check-inexact y 'fpmodulo)
  (fptruncate (C_fmod x y)) )

(define (fpquotient x y)
  (fptruncate (fp/ x y)) )

(define (fpremainder x y)
  (##sys#check-inexact x 'fpremainder)
  (##sys#check-inexact y 'fpremainder)
  (fptruncate (C_remainder x y)) )

;;

(define (fpfraction n)
	(##sys#flonum-fraction n) )

;;;

;;

(define (fp~= x y #!optional (eps flonum-epsilon))
  (let ((diff (fp- x y)))
    (or
      (fpzero? diff)
      (fp<= (fpabs diff) eps) ) ) )

;;;

;;

(define (fpsqr n)
  (fp* n n) )

(define (fpcub n)
  (fp* n (fp* n n)) )

;;;

(define-constant PRECISION-DEFAULT 4.0)

(define-syntax make-unary/precision
  (syntax-rules ()
    ((_ ?op)
      (lambda (n #!optional (p PRECISION-DEFAULT))
        (if (fpzero? p)
          (?op n)
          (let ((precfact (fpprecision-factor p)))
            (fp/ (?op (fp* n precfact)) precfact) ) ) ) ) ) )

;;

(define fptruncate/precision (make-unary/precision fptruncate))
(define fpround/precision (make-unary/precision fpround))
(define fpceiling/precision (make-unary/precision fpceiling))
(define fpfloor/precision (make-unary/precision fpfloor))

;;

(define-constant DEGREE 0.0174532925199432957692369076848861271344) ;pi/180

(define (fpdegree->radian deg)
  (fp* deg DEGREE) )

(define (fpradian->degree rad)
  (fp/ rad DEGREE) )

;;

(define (fpdistance x1 y1 x2 y2)
  (fpsqrt (fpdistance* x1 y1 x2 y2)) )

(define (fpdistance* x1 y1 x2 y2)
  (fp+ (fpsqr (fp- x1 x2)) (fpsqr (fp- y1 y2))) )

;;

(define (fpprecision-factor p #!optional (base 10.0))
  (fpexpt base (exact->inexact p)) )

) ;fp-utils
